import traceback
import typing

from PySide6.QtGui import Qt
from PySide6.QtWidgets import (
    QLabel,
    QLineEdit,
    QPushButton,
    QRadioButton,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from ..guihelper import add_callback_to_standard_signal, build_grid_layout

if typing.TYPE_CHECKING:
    from ..dataexplorer import DataExplorer
    from .datamodel import DataStore

from .dataenums import NUM_TO_CAT_OPS, NumericConversion, NumericConverter


@typing.final
class NumCatWidget(QWidget):
    numeric_converter: NumericConverter

    def __init__(
        self, datastore: "DataStore", column: str, dataexplorer: "DataExplorer"
    ):
        super().__init__()
        self.setStyleSheet(dataexplorer.stylesheet)
        self.debug = dataexplorer.debug
        self.error = dataexplorer.error
        self.column = column
        self.datastore = datastore

        self._layout = QVBoxLayout(self)
        self.setMaximumHeight(200)  # To prevent detaching of title and body.
        column_label = QLabel(column)
        column_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        delete_button = QPushButton("Delete")
        delete_button.setObjectName(
            "DeleteFilter"
        )  # Not the wrong name. Just used for styling.
        _ = delete_button.clicked.connect(self.onDelete)
        self.numeric_converter = NumericConverter(
            conversion=NumericConversion.AS_CATEGORY, value=None
        )
        self.as_category_radio = QRadioButton(NUM_TO_CAT_OPS[0])
        self.as_category_radio.setChecked(True)
        self.bin_n_radio = QRadioButton(NUM_TO_CAT_OPS[1])
        self.bin_n_spinbox = QSpinBox(minimum=1, maximum=1000, value=5)
        self.bin_edges_radio = QRadioButton(NUM_TO_CAT_OPS[2])
        self.bin_edges_line_edit = QLineEdit()
        self.bin_edges_line_edit.setPlaceholderText(
            "Comma-separated list of values: 1, 2, 3"
        )
        build_grid_layout(
            self._layout,
            [
                [column_label],
                [self.as_category_radio],
                [self.bin_n_radio, self.bin_n_spinbox],
                [self.bin_edges_radio, self.bin_edges_line_edit],
                [delete_button],
            ],
        )

        add_callback_to_standard_signal(
            [
                self.as_category_radio,
                self.bin_n_radio,
                self.bin_n_spinbox,
                self.bin_edges_radio,
                self.bin_edges_line_edit,
            ],
            self.on_change,
        )

    def on_change(self):
        if self.as_category_radio.isChecked():
            self.numeric_converter.conversion = NumericConversion.AS_CATEGORY
            self.numeric_converter.value = None
        elif self.bin_n_radio.isChecked():
            self.numeric_converter.conversion = NumericConversion.BINNED
            self.numeric_converter.value = self.bin_n_spinbox.value()
        else:
            self.numeric_converter.conversion = NumericConversion.BIN_WIDTH
            string = self.bin_edges_line_edit.text()
            try:
                list_floats: list[float] = [float(val) for val in string.split(",")]
            except Exception:
                self.error(
                    "The formatting for bin edges is incorrect. Please check it."
                )
                self.debug(traceback.format_exc())
                return
            self.numeric_converter.value = list_floats

        self.datastore.replot_callbacks()

    def onDelete(self):
        self.datastore.on_num_cat_widget_delete(self)
        self.deleteLater()



# pyright: reportUninitializedInstanceVariable=false, reportUnknownMemberType=false, reportMissingTypeStubs=false

import ctypes
import sys
import threading
import typing
from functools import partial
from pathlib import Path
from typing import final

import pandas as pd
from PySide6.QtCore import Qt
from PySide6.QtGui import QCloseEvent, QDragEnterEvent, QDropEvent, QFont, QIcon
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QStatusBar,
    QVBoxLayout,
    QWidget,
)
from qframelesswindow import FramelessMainWindow

from .data.datamodel import DataModel, FilterGUI, NumCatGUI
from .data.importexportGUI import DataImporter, export_data
from .data.tableGUI import TableViewer
from .guihelper import (
    CustomTitleBar,
    build_grid_layout,
    build_layout,
    get_label_widget_row,
)

if typing.TYPE_CHECKING:
    from .dataexplorer import DataExplorer


@final
class DataExplorerGUI(FramelessMainWindow):
    pages: list[QWidget] = []

    def __init__(self, dataexplorer: "DataExplorer"):
        super().__init__()

        self.dataexplorer: DataExplorer = dataexplorer
        self.config = self.dataexplorer.config
        self.debug = self.dataexplorer.debug
        self.error = self.dataexplorer.error
        window_dimensions: list[int] = self.config["General"]["window_dimensions"]

        self.data: dict[str, pd.DataFrame] = {}
        self.clear_message_timeout = None
        self.setWindowTitle("Data Explorer")
        title_bar = CustomTitleBar(self)
        self.setTitle = title_bar.changeTitle
        self.setTitleBar(title_bar)
        self.setMenuWidget(self.titleBar)  # pyright: ignore[reportUnknownArgumentType]
        self.debug(sys.platform)
        if sys.platform == "win32":
            (
                ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(
                    "Data.Explorer"
                )
            )
        self.setWindowIcon(QIcon(self.dataexplorer.icon_path))

        self.setGeometry(100, 100, *window_dimensions)  # x, y, width, height

        self.current_theme = self.config["General"]["default_theme"].casefold()
        self.current_theme_index = self._get_current_theme_index()
        self.debug(self.current_theme)

        # Doesn't really do anything. Most fonts are set by QSS
        QApplication.setFont(QFont(self.dataexplorer.font, self.dataexplorer.font_size))

        self.central_widget = self.dataexplorer.get_widget()
        self.setCentralWidget(self.central_widget)
        self.main_layout = QHBoxLayout(self.central_widget)
        self.main_layout.setContentsMargins(
            0, 0, 0, 0
        )  # No margins for the main layout
        self.main_layout.setSpacing(0)  # No spacing between sidebar and content
        statusbar = self.statusBar()
        if not statusbar:
            self.debug("Unable to load status bar!")
        else:
            self.QStatusBar: QStatusBar = statusbar

        self.content_area_container = QFrame()
        self.content_area_container.setObjectName("ContentArea")
        self.content_layout = QVBoxLayout(self.content_area_container)

        self._generate_pages()
        self._generate_sidebar()

        # Add sidebar and content area to the main layout
        self.main_layout.addWidget(self.sidebar)
        self.main_layout.addWidget(self.content_area_container, 1)

        self.set_theme()
        self.write_to_status_bar("Initialisation complete!")

    def _generate_sidebar(self):
        self.sidebar = QFrame()
        self.sidebar.setObjectName("Sidebar")  # For QSS styling
        self.sidebar.setFixedWidth(240)  # Adjust width as needed
        self.sidebar_layout = QVBoxLayout(self.sidebar)
        self.sidebar_layout.setContentsMargins(0, 0, 0, 0)
        self.sidebar_layout.setSpacing(5)  # Spacing between sidebar items
        self.sidebar_layout.addSpacerItem(
            QSpacerItem(20, 20, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        )

        idx: int = 0

        # Navigation Buttons
        self.nav_button1 = QPushButton("Manage Data")
        self.nav_button1.setCheckable(True)  # Allows it to stay "pressed"
        self.nav_button1.setChecked(True)  # Default selected
        f = partial(self.switch_content_page, index=idx)
        _ = self.nav_button1.clicked.connect(f)
        idx += 1

        self.nav_button2 = QPushButton("Filter Data")
        self.nav_button2.setCheckable(True)
        f = partial(self.switch_content_page, index=idx)
        _ = self.nav_button2.clicked.connect(f)
        idx += 1

        self.nav_button3 = QPushButton("Convert Numeric to Categorical")
        self.nav_button3.setCheckable(True)
        f = partial(self.switch_content_page, index=idx)
        _ = self.nav_button3.clicked.connect(f)
        idx += 1

        self.nav_button4 = QPushButton("Plot Data")
        self.nav_button4.setCheckable(True)
        f = partial(self.switch_content_page, index=idx)
        _ = self.nav_button4.clicked.connect(f)
        idx += 1

        self.nav_button5 = QPushButton("Plot Settings")
        self.nav_button5.setCheckable(True)
        f = partial(self.switch_content_page, index=idx)
        _ = self.nav_button5.clicked.connect(f)
        idx += 1

        self.nav_button6 = QPushButton("App Settings")
        self.nav_button6.setCheckable(True)
        f = partial(self.switch_content_page, index=idx)
        _ = self.nav_button6.clicked.connect(f)
        idx += 1

        self.sidebar_nav_buttons = [
            self.nav_button1,
            self.nav_button2,
            self.nav_button3,
            self.nav_button4,
            self.nav_button5,
            self.nav_button6,
        ]

        for btn_name in self.plugin_button_names:
            btn = QPushButton(btn_name)
            btn.setCheckable(True)
            f = partial(self.switch_content_page, index=idx)
            _ = btn.clicked.connect(f)
            self.sidebar_nav_buttons.append(btn)
            idx = idx + 1
            self.debug(btn)

        build_layout(self.sidebar_layout, self.sidebar_nav_buttons)

        # Spacer to push nav items to the top
        self.sidebar_layout.addSpacerItem(
            QSpacerItem(
                20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
            )
        )

    def _generate_pages(self):
        # Page 1: Data Import/Export/View
        self.pages.append(self._manage_data_page())

        # Page 2: Filtering
        self.pages.append(self._filter_data_page())

        # Page 3: Numeric -> Categorical
        self.pages.append(self._num_to_cat_page())

        # Page 4: Plotting
        self.pages.append(self._plotting_page())

        # Page 5: Plot Settings
        self.pages.append(self.plot_settings_page())

        # Page 6: App Settings
        self.pages.append(self.app_settings_page())

        self.pages.extend(self.load_plugin_pages())

        # Add initial page to content layout
        for page_widget in self.pages:
            self.content_layout.addWidget(page_widget)
            page_widget.setVisible(False)

        # Show the first page
        if self.pages:
            self.pages[0].setVisible(True)
            self.current_page_index = 0

    def _manage_data_page(self) -> QWidget:
        manage_data_page = self.dataexplorer.get_widget()
        manage_data_page_layout = QVBoxLayout(manage_data_page)
        manage_data_page.setObjectName("ContentArea")

        top_spacer = QSpacerItem(
            10, 30, QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed
        )

        import_button = QPushButton("Import Data from File")
        _ = import_button.clicked.connect(self._on_import_data_from_file)

        mid_spacer = QSpacerItem(
            10, 10, QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed
        )

        active_dataset_label = QLabel("Set Active Dataset")
        active_dataset_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        mid_spacer_2 = QSpacerItem(
            10, 10, QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed
        )

        self.active_dataset_combobox = QComboBox()
        self.active_dataset_combobox.setEditable(True)

        le = self.active_dataset_combobox.lineEdit()
        if le is None:
            self.debug("Unable to get lineEdit property of QComboBox.")
        else:
            le.setAlignment(Qt.AlignmentFlag.AlignCenter)
            le.setReadOnly(True)
        _ = self.active_dataset_combobox.currentTextChanged.connect(
            self.set_active_dataset
        )

        mid_spacer_3 = QSpacerItem(
            10, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
        )

        self.filter_viewexport_checkbox = QCheckBox(
            "Apply filter for exporting/viewing"
        )
        self.filter_viewexport_checkbox.setChecked(True)

        export_data_button = QPushButton("Export Dataset")
        _ = export_data_button.clicked.connect(self._export_data)

        bottom_spacer = QSpacerItem(
            10, 10, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed
        )

        view_data_button = QPushButton("View Dataset")
        _ = view_data_button.clicked.connect(self._view_data)

        view_data_statistics_button = QPushButton("View Data Statistics")
        _ = view_data_statistics_button.clicked.connect(self._view_data_statistics)
        widg_list = [
            top_spacer,
            import_button,
            mid_spacer,
            active_dataset_label,
            mid_spacer_2,
            self.active_dataset_combobox,
            mid_spacer_3,
            self.filter_viewexport_checkbox,
            export_data_button,
            bottom_spacer,
            view_data_button,
            bottom_spacer,
            view_data_statistics_button,
        ]

        build_layout(manage_data_page_layout, widg_list)
        manage_data_page_layout.addStretch()
        manage_data_page.setAcceptDrops(True)
        manage_data_page.dragEnterEvent = self.manage_data_dragEnterEvent
        manage_data_page.dropEvent = self.manage_data_dropEvent
        return manage_data_page

    def manage_data_dragEnterEvent(self, event: QDragEnterEvent):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def manage_data_dropEvent(self, event: QDropEvent):
        if event.mimeData().hasUrls():
            file_path = event.mimeData().urls()[0].toLocalFile()
            event.accept()
            if Path(file_path).exists():
                self._importer = DataImporter(self.dataexplorer, file_path)
            else:
                self.error("That file did not work! It does not exist.")
        else:
            event.ignore()

    def _view_data(self):
        datamodel: DataModel = self.dataexplorer.datamodel
        if self.filter_viewexport_checkbox.isChecked():
            active = datamodel.active_filtered_data
        else:
            active = datamodel.active_cleaned_data
        if active is None:
            self.write_to_status_bar("No Active Data to view")
        else:
            self.table_viewer = TableViewer(self.dataexplorer, active, datamodel.active)

    def _view_data_statistics(self):
        datamodel: DataModel = self.dataexplorer.datamodel
        if self.filter_viewexport_checkbox.isChecked():
            active = datamodel.active_filtered_data
        else:
            active = datamodel.active_cleaned_data
        if active is None:
            self.write_to_status_bar("No Active Data to view")
        else:
            active = active.describe(include="all")
            self.table_viewer = TableViewer(
                self.dataexplorer, active, f"{datamodel.active} Statistics"
            )

    def _on_import_data_from_file(self):
        self._importer = DataImporter(self.dataexplorer)

    def set_active_dataset(self):
        self.dataexplorer.datamodel.set_active_dataset(
            self.active_dataset_combobox.currentText()
        )
        for dataset in self.filter_pages:
            self.filter_pages[dataset].setVisible(False)
            self.num_to_cat_pages[dataset].setVisible(False)
        self.filter_pages[self.dataexplorer.datamodel.active].setVisible(True)
        self.num_to_cat_pages[self.dataexplorer.datamodel.active].setVisible(True)
        self.setTitle(f"Current dataset: {self.dataexplorer.datamodel.active}")
        self.debug("Active dataset changed to: " + self.dataexplorer.datamodel.active)

    def update_data_list(self, name: str):
        datalist: QComboBox = self.active_dataset_combobox
        datalist.addItem(name)

    def _export_data(self):
        datamodel: DataModel = self.dataexplorer.datamodel
        if self.filter_viewexport_checkbox.isChecked():
            active = datamodel.active_filtered_data
        else:
            active = datamodel.active_cleaned_data
        if active is None:
            self.write_to_status_bar("No Active Data to view")
        else:
            export_data(self.dataexplorer, active)

    def _filter_data_page(self) -> QFrame:
        filter_data_page = QFrame()
        self.filter_data_page_layout = QVBoxLayout(filter_data_page)
        self.filter_pages: dict[str, QWidget] = {}
        self.filter_pages["__empty__"] = QLabel("No active datasets to filter")
        self.filter_data_page_layout.addWidget(self.filter_pages["__empty__"])
        return filter_data_page

    def add_filter_GUI_to_page(self, filterGUI: FilterGUI, name: str):
        self.filter_data_page_layout.addWidget(filterGUI.filter_page)
        filterGUI.filter_page.setObjectName("ContentArea")
        self.filter_pages[name] = filterGUI.filter_page
        self.filter_pages[name].setVisible(False)

    def _num_to_cat_page(self) -> QFrame:
        num_to_cat_page = QFrame()
        self.num_to_cat_page_layout = QVBoxLayout(num_to_cat_page)
        self.num_to_cat_pages: dict[str, QWidget] = {}
        self.num_to_cat_pages["__empty__"] = QLabel("No active datasets")
        self.num_to_cat_page_layout.addWidget(self.num_to_cat_pages["__empty__"])
        return num_to_cat_page

    def add_num_cat_GUI_to_page(self, numCatGUI: NumCatGUI, name: str):
        self.num_to_cat_page_layout.addWidget(numCatGUI.num_to_cat_page)
        numCatGUI.num_to_cat_page.setObjectName("ContentArea")
        self.num_to_cat_pages[name] = numCatGUI.num_to_cat_page
        self.num_to_cat_pages[name].setVisible(False)

    def _plotting_page(self) -> QWidget:
        plotting_page = self.dataexplorer.get_widget()
        plotting_page.setObjectName("ContentArea")
        plotting_page_layout = QVBoxLayout(plotting_page)

        histogram_button = QPushButton("Histogram")
        _ = histogram_button.clicked.connect(
            lambda: self.dataexplorer.plotter.histogram_plotter()
        )

        scatter_plot_button = QPushButton("Scatter")
        _ = scatter_plot_button.clicked.connect(
            lambda: self.dataexplorer.plotter.scatterplot_plotter()
        )
        line_plot_button = QPushButton("Line")
        _ = line_plot_button.clicked.connect(
            lambda: self.dataexplorer.plotter.lineplot_plotter()
        )

        categorical_plot_button = QPushButton("Categorical")
        _ = categorical_plot_button.clicked.connect(
            lambda: self.dataexplorer.plotter.catplot_plotter()
        )
        count_plot_button = QPushButton("Categorical Count")
        _ = count_plot_button.clicked.connect(
            lambda: self.dataexplorer.plotter.countplot_plotter()
        )

        correl_matrix = QPushButton("Correlation Matrix")
        _ = correl_matrix.clicked.connect(
            lambda: self.dataexplorer.plotter.correlmatrix_plotter()
        )

        regression_plot_button = QPushButton("Linear Regression")
        _ = regression_plot_button.clicked.connect(
            lambda: self.dataexplorer.plotter.regression_plotter()
        )

        pair_grid_button = QPushButton("Pair Grid")
        _ = pair_grid_button.clicked.connect(
            lambda: self.dataexplorer.plotter.pair_grid_plotter()
        )

        build_grid_layout(
            plotting_page_layout,
            [
                [histogram_button, scatter_plot_button],
                [line_plot_button, categorical_plot_button],
                [count_plot_button, correl_matrix],
                [regression_plot_button, pair_grid_button],
            ],
        )

        plotting_page_layout.addStretch()

        return plotting_page

    def plot_settings_page(self) -> QWidget:
        plot_settings_page = self.dataexplorer.get_widget()
        plot_settings_page.setObjectName("ContentArea")
        self.plot_settings_layout = QVBoxLayout(plot_settings_page)
        return plot_settings_page

    def app_settings_page(self) -> QWidget:
        app_settings_page = self.dataexplorer.get_widget()
        app_settings_page.setObjectName("ContentArea")
        app_settings_layout = QVBoxLayout(app_settings_page)
        self.theme_combobox = QComboBox()
        self.theme_combobox.addItems(self.config["Themes"].keys())
        self.theme_combobox.setCurrentIndex(self.current_theme_index)
        _ = self.theme_combobox.currentTextChanged.connect(self.set_theme)
        theme_combobox = get_label_widget_row("Theme", self.theme_combobox)
        build_layout(app_settings_layout, [theme_combobox])
        app_settings_layout.addStretch()
        return app_settings_page

    def write_to_status_bar(self, message: str):
        if self.clear_message_timeout is not None:
            self.clear_message_timeout.cancel()
        self.QStatusBar.showMessage(message)
        self.clear_message_timeout = threading.Timer(30, self.clear_status_bar)
        self.clear_message_timeout.start()

    def clear_status_bar(self):
        self.QStatusBar.clearMessage()
        self.clear_message_timeout = None

    def _get_current_theme_index(self) -> int:
        for i, theme_name in enumerate(self.config["Themes"].keys()):
            if self.current_theme.casefold() == theme_name.casefold():
                return i
        self.debug("Could not get current theme index!")
        return 0

    def set_theme(self):
        self.current_theme = self.theme_combobox.currentText()
        for theme_name in self.config["Themes"].keys():
            if self.current_theme.casefold() == theme_name.casefold():
                self.dataexplorer.stylesheet = self.config["Themes"][theme_name]
                self.debug(self.dataexplorer.stylesheet)
                self.setStyleSheet(self.dataexplorer.stylesheet)
                break

    def load_plugin_pages(self) -> list[QWidget]:
        plugin_loaders = self.dataexplorer.plugin_list
        self.plugin_button_names: list[str] = []
        plugin_widgets: list[QWidget] = []
        for plugin_loader in plugin_loaders:
            widg, btn_name = plugin_loader(self.dataexplorer)
            self.plugin_button_names.append(btn_name)
            plugin_widgets.append(widg)

        return plugin_widgets

    def switch_content_page(self, index: int):
        self.debug(f"Switch to page {index}")
        if 0 <= index < len(self.pages):
            self.pages[self.current_page_index].setVisible(False)
            self.sidebar_nav_buttons[self.current_page_index].setChecked(False)
            self.pages[index].setVisible(True)
            self.sidebar_nav_buttons[index].setChecked(True)
            self.current_page_index = index

    @typing.override
    def closeEvent(self, event: QCloseEvent):
        if self.clear_message_timeout is not None:
            self.clear_message_timeout.cancel()
        self.dataexplorer.closeEvent()
        return super().closeEvent(event)

    def _not_implemented(self):
        _ = QMessageBox.warning(
            self, "Not Implemented", "This functionality has not yet been implemented."
        )

    def error_msg(self, message: str):
        _ = QMessageBox.warning(self, "Error", message)



# pyright: reportUnknownVariableType=false, reportUninitializedInstanceVariable=false
import typing
from collections.abc import Iterable
from functools import partial
from pathlib import Path

import pandas as pd
from PySide6.QtCore import (
    Qt,
)
from PySide6.QtGui import QFont
from PySide6.QtWidgets import (
    QComboBox,
    QFileDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QRadioButton,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)

from ..guihelper import build_layout, get_dynamic_scroll_area, get_label_widget_row
from .dataenums import (
    EXPORT_FILE_TYPES,
    IMPORT_FILE_TYPES,
    IMPORT_FILE_TYPES_LIST,
    NAN_CAT,
    NAN_DATETIME,
    NAN_NUM,
    VALID_DTYPES,
    NaNOperation,
)
from .datamodel import (
    DataModel,
    categorical_comparator,
    datetime_comparator,
    handle_dtype_operation,
    handle_nan_operation,
    numeric_comparator,
)

if typing.TYPE_CHECKING:
    from ..dataexplorer import DataExplorer


@typing.final
class DataImporter:
    data: pd.DataFrame

    def __init__(self, dataexplorer: "DataExplorer", file_path: str = ""):
        self.dataexplorer: "DataExplorer" = dataexplorer
        self.datamodel: DataModel = dataexplorer.datamodel
        self.error = dataexplorer.error
        self.debug = dataexplorer.debug

        filetype: str
        if file_path == "":
            file_path, _ = QFileDialog.getOpenFileName(
                parent=None,
                caption="Open Data File",
                filter=";;".join(IMPORT_FILE_TYPES),
            )
        else:
            filetype = Path(file_path).suffix
            self.debug(filetype)
            if filetype not in "".join(IMPORT_FILE_TYPES_LIST):
                self.error("Wrong file type")
                return
        if not file_path:
            self.error("Empty File Path.")
            return

        self.dataexplorer.info(f"File Selected: {file_path}")
        self.file_name: str = Path(file_path).stem
        filetype = Path(file_path).suffix
        self._read_file(file_path, filetype)

    def _validate_import(self):
        self.debug(f"{self.data.head()}")
        self.debug(f"{self.data.shape}")
        self._validate_import_widget = self._create_widget(800, 600)
        self.layout = QVBoxLayout(self._validate_import_widget)
        self.layout.setSpacing(5)

        top_spacer = QSpacerItem(20, 20)

        top_label = QLabel("Data Details")
        top_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        shape_label = QLabel(
            f"Rows: {self.data.shape[0]},\nColumns: {self.data.shape[1]}"
        )
        shape_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        column_label = QLabel("Column Names:\n")
        column_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        columns_scroll_area, scroll_area_widget = get_dynamic_scroll_area(
            self.dataexplorer.get_widget, width=750
        )
        scroll_area_vbox = QVBoxLayout(scroll_area_widget)
        scroll_area_layout = QGridLayout()

        for i, column in enumerate(self.data.columns):
            label = QLabel(column)  # pyright: ignore[reportUnknownArgumentType]
            scroll_area_layout.addWidget(label, self._row(i), self._col(i))
        build_layout(scroll_area_vbox, [column_label, scroll_area_layout])
        nan_label = QLabel("Default operation for missing values")

        nan_hbox = QHBoxLayout()
        nan_hbox.setSpacing(5)
        nan_sub_vbox = QVBoxLayout()

        categorical_hbox = QHBoxLayout()
        label_nan_categorical = QLabel("Categorical: ")
        self.default_nan_categorical: QComboBox = QComboBox()
        build_layout(
            categorical_hbox, [label_nan_categorical, self.default_nan_categorical]
        )

        numeric_hbox = QHBoxLayout()
        label_nan_numeric = QLabel("Numeric: ")
        self.default_nan_numeric: QComboBox = QComboBox()
        build_layout(numeric_hbox, [label_nan_numeric, self.default_nan_numeric])

        datetime_hbox = QHBoxLayout()
        label_nan_datetime = QLabel("Datetime: ")
        self.default_nan_datetime: QComboBox = QComboBox()
        build_layout(datetime_hbox, [label_nan_datetime, self.default_nan_datetime])

        build_layout(nan_sub_vbox, [categorical_hbox, numeric_hbox, datetime_hbox])
        build_layout(nan_hbox, [nan_label, nan_sub_vbox])

        self.default_nan_categorical.addItems(NAN_CAT)
        self.default_nan_numeric.addItems(NAN_NUM)
        self.default_nan_datetime.addItems(NAN_DATETIME)

        continue_button = QPushButton("Continue to set column data types")
        _ = continue_button.clicked.connect(self._set_dtypes)

        self.default_custom_radio_button = (
            QRadioButton(
                "Handle missing values by column data type "
                "using defaults (defaults set above)"
            ),
            QRadioButton("Handle missing values on a per-column basis"),
        )
        self.default_custom_radio_button[1].setChecked(True)
        radio_button_vbox = QVBoxLayout()
        build_layout(radio_button_vbox, [*self.default_custom_radio_button])

        widget_list = [
            top_spacer,
            top_label,
            shape_label,
            columns_scroll_area,
            nan_hbox,
            radio_button_vbox,
            continue_button,
        ]
        build_layout(self.layout, widget_list)
        self._validate_import_widget.setLayout(self.layout)
        self._validate_import_widget.update()
        self._validate_import_widget.show()

    def _set_dtypes(self):
        self.default_nan_cat: str = self.default_nan_categorical.currentText()
        self.default_nan_num: str = self.default_nan_numeric.currentText()
        self.default_nan_dt: str = self.default_nan_datetime.currentText()
        self._close_widget(self._validate_import_widget)

        self._dtype_widget = self._create_widget(800, 600)
        self.dtypes: dict[str, str] = {}
        self.layout = QVBoxLayout()

        top_spacer = QSpacerItem(20, 20)
        dtype_grid = QGridLayout()
        self.dtype_widgets: dict[str, QComboBox] = {}

        for i, column in enumerate(self.data.columns):
            try:
                data_column = self.data[column]
                assert isinstance(data_column, pd.Series)
                if numeric_comparator(data_column):
                    widg = self.dtype_widgets[column] = QComboBox()
                    widg.addItems(VALID_DTYPES)
                    widg.setCurrentText(VALID_DTYPES[1])
                elif categorical_comparator(data_column):
                    widg = self.dtype_widgets[column] = QComboBox()
                    widg.addItems(VALID_DTYPES)
                    widg.setCurrentText(VALID_DTYPES[0])
                elif datetime_comparator(data_column):
                    widg = self.dtype_widgets[column] = QComboBox()
                    widg.addItems(VALID_DTYPES)
                    widg.setCurrentText(VALID_DTYPES[2])
                else:
                    self.error(f"Invalid dtype seen. {column}")
                    self._close_widget(self._dtype_widget)
                    return
                function = partial(self._dtype_setter, column=column)  # pyright: ignore[reportUnkownMemberType]
                _ = widg.currentTextChanged.connect(function)
                hbox = get_label_widget_row(f"{column}:", widg)
                dtype_grid.addLayout(hbox, self._row(i), self._col(i))
                self.debug(
                    f'Dtype, Row, Col: "{widg.currentText()}", '
                    f"{self._row(i)}, {self._col(i)}"
                )
            except TypeError as e:
                self.error(f"Error reading column {column}: {e}")
                self._close_widget(self._dtype_widget)
                return
            except AssertionError as e:
                self.error(f"self.data[{column}] did not return a series {e}")
                self._close_widget(self._dtype_widget)
                return

        button = QPushButton("Continue")
        _ = button.clicked.connect(self._process_dtypes)
        scroll_content_widget, content_widget = get_dynamic_scroll_area(
            self.dataexplorer.get_widget, width=750
        )
        content_widget.setLayout(dtype_grid)
        build_layout(self.layout, [top_spacer, scroll_content_widget, button])
        self._dtype_widget.setLayout(self.layout)
        self._dtype_widget.show()

    def _dtype_setter(self, *value, column: str):
        self.debug(f"{column} ComboBox Changed")
        self.dtypes[column] = value[0]

    def _process_dtypes(self):
        for column in self.dtypes:
            self.data[column] = handle_dtype_operation(
                self.data.loc[:, column], self.dtypes[column], self.debug
            )
            self.debug(
                f"Dtype conversion applied: {self.dtypes[column]} to column {column}"
            )
        self._handle_nans()

    def _row(self, n: int) -> int:
        grid_max_col = self.dataexplorer.config["Grid"]["grid_max_col"]
        return int(n / grid_max_col)

    def _col(self, n: int) -> int:
        grid_max_col = self.dataexplorer.config["Grid"]["grid_max_col"]
        return n % grid_max_col

    def _handle_nans(self):
        self._close_widget(self._dtype_widget)
        self._nan_widget = self._create_widget(800, 600)
        self.layout = QVBoxLayout()

        label = QLabel("Handle Missing Values")
        label.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        nan_grid = QGridLayout()
        self.nan_widgets = {}

        top_spacer = QSpacerItem(20, 20)
        nan_columns: Iterable[str] = self.data.columns[self.data.isnull().any()]

        if len(nan_columns) == 0:
            self._process_nans()
            return

        for i, column in enumerate(nan_columns):
            try:
                data_column = self.data[column]
                assert isinstance(data_column, pd.Series)
                self.debug(f"{column}: {data_column.dtype}")
                if numeric_comparator(data_column):
                    widg = self.nan_widgets[column] = QComboBox()
                    widg.addItems(NAN_NUM)
                    widg.setCurrentText(self.default_nan_num)
                elif categorical_comparator(data_column):
                    widg = self.nan_widgets[column] = QComboBox()
                    widg.addItems(NAN_CAT)
                    widg.setCurrentText(self.default_nan_cat)
                elif datetime_comparator(data_column):
                    widg = self.nan_widgets[column] = QComboBox()
                    widg.addItems(NAN_DATETIME)
                    widg.setCurrentText(self.default_nan_dt)
                else:
                    self.error(f"Invalid dtype seen {column}.")
                    self._close_widget(self._nan_widget)
                    return
                hbox = get_label_widget_row(f"{column}:", widg)
                nan_grid.addLayout(hbox, self._row(i), self._col(i))
                self.debug(
                    f'NaN, Row, Col: "{widg.currentText()}", '
                    f"{self._row(i)}, {self._col(i)}"
                )
            except TypeError as e:
                self.error(f"Error reading column {column}: {e}")
                self._close_widget(self._nan_widget)
                return
            except AssertionError as e:
                self.error(f"self.data[{column}] did not return a series {e}")
                self._close_widget(self._dtype_widget)
                return

        if self.default_custom_radio_button[0].isChecked():
            self._process_nans()
            return

        nan_grid.setAlignment(Qt.AlignmentFlag.AlignCenter)
        button = QPushButton("Continue")
        _ = button.clicked.connect(self._process_nans)
        scroll_content_widget, content_widget = get_dynamic_scroll_area(
            self.dataexplorer.get_widget, width=750
        )
        content_widget.setLayout(nan_grid)
        build_layout(self.layout, [top_spacer, label, scroll_content_widget, button])
        self._nan_widget.setLayout(self.layout)
        self._nan_widget.show()

    def _process_nans(self):
        self.nans = {}
        n_rows = self.data.shape[0]
        for column in self.nan_widgets:
            self.nans[column] = self.nan_widgets[column].currentText()
            self.data = handle_nan_operation(
                self.data, self.nans[column], column, self.debug
            )
            if self.nans[column] == NaNOperation.DROP_ROWS.value:
                self.debug(
                    f"Dropped {n_rows - self.data.shape[0]} rows for column {column}"
                )
                n_rows = self.data.shape[0]
        self._close_widget(self._nan_widget)
        self._add_data()

    def _add_data(self):
        self.debug(self.nans)
        self.debug(self.dtypes)
        self.datamodel.add_dataset_to_model(
            self.data, self.file_name, {"NaN": self.nans, "dtype": self.dtypes}
        )

    def _create_widget(self, w: int, h: int) -> QWidget:
        widget = self.dataexplorer.get_widget(detached=True)
        widget.setObjectName("DataImporter")
        widget.setWindowTitle("Data Importer")
        widget.resize(w, h)
        widget.setFont(QFont(self.dataexplorer.font, self.dataexplorer.font_size))
        self.dataexplorer.owned_widgets.append(widget)
        return widget

    def _read_file(self, file_path: str, filetype: str):
        if filetype in IMPORT_FILE_TYPES_LIST[0:3]:
            excel_file = pd.ExcelFile(file_path)
            if len(excel_file.sheet_names) > 1:
                self.select_worksheet_widget = self._create_widget(200, 300)
                layout = QVBoxLayout(self.select_worksheet_widget)
                self._sheet_combobox = QComboBox()
                sheet_names = [str(name) for name in excel_file.sheet_names]
                self._sheet_combobox.addItems(sheet_names)
                sheet_combobox = get_label_widget_row(
                    "Select Worksheet:", self._sheet_combobox
                )
                button = QPushButton("Continue")
                button.clicked.connect(lambda: self._import_data_excel(file_path))
                build_layout(layout, [sheet_combobox, button])
                self.select_worksheet_widget.show()
            else:
                self.data = pd.read_excel(file_path, engine="calamine")
                self._validate_import()
        elif filetype == IMPORT_FILE_TYPES_LIST[3]:
            self.data = pd.read_csv(file_path)
            self._validate_import()
        elif filetype == IMPORT_FILE_TYPES_LIST[4]:
            self.data = pd.read_parquet(file_path)
            self._validate_import()
        elif filetype == IMPORT_FILE_TYPES_LIST[5]:
            self.data = pd.read_feather(file_path)
            self._validate_import()
        else:
            self.error("Import failed! Invalid filetype.")

    def _import_data_excel(self, file_path: str):
        sheet = self._sheet_combobox.currentText()
        self.data = pd.read_excel(file_path, sheet_name=sheet, engine="calamine")
        self.file_name = self.file_name + f" Sheet:{sheet}"
        self.select_worksheet_widget.close()
        self._validate_import()

    def _close_widget(self, widget: QWidget):
        self.dataexplorer.owned_widgets.remove(widget)
        _ = widget.close()


def export_data(dataexplorer: "DataExplorer", data: pd.DataFrame):
    file_path, selFilter = QFileDialog.getSaveFileName(
        parent=None,
        caption="Save Data File",
        filter=";;".join(EXPORT_FILE_TYPES),
    )
    if file_path:
        try:
            if selFilter.startswith("Excel"):
                data = get_naive_dataframe(data)
                data.to_excel(file_path, index=True)
            elif selFilter.startswith("CSV"):
                data.to_csv(file_path, index=True)
            elif selFilter.startswith("Parquet"):
                data.to_parquet(file_path, index=True)
            elif selFilter.startswith("Feather"):
                data.to_feather(file_path)
        except Exception as e:
            dataexplorer.error(str(e))
        else:
            dataexplorer.status_message("File saved")
    else:
        dataexplorer.status_message("No save file set.")


def get_naive_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    for col in df.columns:
        try:
            df[col] = df[col].dt.tz_localize(None)
        except Exception:
            pass
    try:
        df.index = df.index.tz_localize(None)  # pyright: ignore[reportAttributeAccessIssue]
    except Exception:
        pass
    return df



import pandas as pd
from PySide6.QtCore import QAbstractTableModel, Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QSpacerItem, QTableView, QVBoxLayout


class TableViewer:
    def __init__(self, dataexplorer, data: pd.DataFrame, name: str):
        self.dataexplorer = dataexplorer
        self.table_viewer = self.dataexplorer.get_widget(detached=True)
        self.table_viewer.setObjectName("TableViewer")
        self.table_viewer.setWindowTitle(f"Table Viewer: {name}")
        self.table_viewer.setGeometry(100, 100, 800, 400)
        self.table_viewer.setFont(QFont(dataexplorer.font, dataexplorer.font_size))
        self.table_view = QTableView()
        self.model = TableViewModel(data)
        self.table_view.setModel(self.model)
        layout = QVBoxLayout()
        layout.addSpacerItem(QSpacerItem(20, 20))
        layout.addWidget(self.table_view)
        self.table_viewer.setLayout(layout)
        self.table_viewer.show()


class TableViewModel(QAbstractTableModel):
    def __init__(self, data: pd.DataFrame):
        super().__init__()
        self._data = data

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, parent=None):
        return self._data.shape[1]

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole:
                return str(self._data.iloc[index.row(), index.column()])
        return None

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return str(self._data.columns[section])
            elif orientation == Qt.Orientation.Vertical:
                return str(self._data.index[section])
        return None



import typing
from dataclasses import dataclass

import pandas as pd
from PySide6.QtCore import QDateTime, Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QDateTimeEdit,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
from superqt import QLabeledDoubleRangeSlider

from ..guihelper import MultiSelectComboBox, build_layout

if typing.TYPE_CHECKING:
    from ..dataexplorer import DataExplorer
    from .datamodel import DataStore
from .dataenums import Dtype


@dataclass
class FilterStore:
    dtype: Dtype
    filter_value: tuple[float, float] | tuple[pd.Timestamp, pd.Timestamp] | list[str]
    active: bool = True


class InvalidFilterWidgetError(RuntimeError):
    pass


@typing.final
class FilterWidget(QWidget):
    filterstore: FilterStore

    def __init__(
        self, datastore: "DataStore", column: str, dataexplorer: "DataExplorer"
    ):
        super().__init__()
        self.setStyleSheet(dataexplorer.stylesheet)
        self.debug = dataexplorer.debug
        self.error = dataexplorer.error
        self.dtype = column_type = datastore.get_column_dtype(column)
        self.column = column
        self.datastore = datastore

        self._layout = QVBoxLayout()
        self.setMaximumHeight(150)  # To prevent detaching of title and filter.
        self.setMinimumWidth(275)  # To prevent squishing of certain filters

        self.column_label = QLabel(column)
        self.column_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        self.column_label.setObjectName("FilterWidgetLabel")
        self.active_checkbox = QCheckBox()
        self.active_checkbox.setChecked(True)
        _ = self.active_checkbox.toggled.connect(lambda: self.on_state_change())
        header_layout = QHBoxLayout()
        header_layout.addWidget(self.active_checkbox)
        header_layout.addWidget(self.column_label, stretch=1)

        delete_button = QPushButton("Delete Filter")
        delete_button.setObjectName("DeleteFilter")
        _ = delete_button.clicked.connect(self.onDelete)
        match column_type:
            case Dtype.NUMERIC:
                minimum, maximum = (  # pyright: ignore[reportUnknownVariableType]
                    datastore.cleaned_data[column].min(),
                    datastore.cleaned_data[column].max(),
                )
                if maximum == minimum:
                    self.deleteLater()
                    self.debug("Unable to create a filter widget for this.")
                    self.error(
                        f"Unable to create filter widget for {column} because max=min"
                    )
                    raise InvalidFilterWidgetError()

                self.filter_widget = QLabeledDoubleRangeSlider()
                self.filter_widget.setObjectName("LabeledRangeSlider")
                self.debug(f"min: {minimum} max: {maximum}")
                self.filter_widget.setRange(minimum, maximum)  # noqa # pyright: ignore[reportUnknownArgumentType]
                self.filter_widget.setValue((minimum, maximum))
                self.filter_widget._min_label.setMaximumWidth(40)
                self.filter_widget._max_label.setMaximumWidth(40)
                # self.filter_widget.setEdgeLabelMode(
                #     QLabeledDoubleRangeSlider.EdgeLabelMode.LabelIsValue
                # )
                _ = self.filter_widget.valueChanged.connect(self.on_numeric_change)
                self.filterstore = FilterStore(self.dtype, (minimum, maximum))  # pyright: ignore[reportUnknownArgumentType]
                build_layout(
                    self._layout, [header_layout, self.filter_widget, delete_button]
                )
                # TODO: Figure out a better workaround for this code.
                self.filter_widget._min_label.editingFinished.emit()
                self.filter_widget._max_label.editingFinished.emit()
            case Dtype.CATEGORICAL:
                self.filter_widget = MultiSelectComboBox()
                self.filter_widget.addItems(  # pyright: ignore[reportUnknownMemberType]
                    list(datastore.cleaned_data[column].astype(str).unique())
                )
                _ = self.filter_widget.dataChanged.connect(self.on_categorical_change)
                self.filterstore = FilterStore(
                    self.dtype, list(datastore.cleaned_data[column].unique())
                )
                build_layout(
                    self._layout, [header_layout, self.filter_widget, delete_button]
                )
            case Dtype.DATETIME:
                minimum, maximum = (
                    datastore.cleaned_data[column].min(),
                    datastore.cleaned_data[column].max(),
                )
                if not isinstance(maximum, pd.Timestamp) or not isinstance(
                    minimum, pd.Timestamp
                ):
                    self.deleteLater()
                    self.debug(f"Datetime values are of wrong type for column {column}")
                    self.error(
                        f"Unable to create a filter widget for {column}"
                        "due to a type issue."
                    )
                    raise InvalidFilterWidgetError()

                def get_time(x: pd.Timestamp) -> QDateTime:
                    return QDateTime(x.year, x.month, x.day, x.hour, x.minute, x.second)

                hbox_min = QHBoxLayout()
                min_value_label = QLabel("Min")
                self.min_value = QDateTimeEdit()
                self.min_value.setMinimumDateTime(get_time(minimum))
                self.min_value.setMaximumDateTime(get_time(maximum))
                self.min_value.setDateTime(get_time(minimum))
                self.min_value.setCalendarPopup(True)
                _ = self.min_value.dateTimeChanged.connect(self.on_datetime_change)
                build_layout(hbox_min, [min_value_label, self.min_value])

                hbox_max = QHBoxLayout()
                max_value_label = QLabel("Max")
                self.max_value = QDateTimeEdit()
                self.max_value.setMinimumDateTime(get_time(minimum))
                self.max_value.setMaximumDateTime(get_time(maximum))
                self.max_value.setDateTime(get_time(maximum))
                self.max_value.setCalendarPopup(True)
                _ = self.max_value.dateTimeChanged.connect(self.on_datetime_change)
                build_layout(hbox_max, [max_value_label, self.max_value])

                self.filterstore = FilterStore(self.dtype, (minimum, maximum))

                build_layout(
                    self._layout, [header_layout, hbox_min, hbox_max, delete_button]
                )
        self.setLayout(self._layout)
        self.update_label_style()

    def on_state_change(self):
        self.filterstore.active = self.active_checkbox.isChecked()
        self.debug(f"{self.filterstore} state toggled")
        self.update_label_style()

    def update_label_style(self):
        _ = self.column_label.setProperty("active", self.filterstore.active)
        self.column_label.style().unpolish(self.column_label)
        self.column_label.style().polish(self.column_label)

    def on_numeric_change(self):
        if isinstance(self.filter_widget, QLabeledDoubleRangeSlider):
            self.filterstore.filter_value = self.filter_widget.value()
            self.debug(self.filterstore.filter_value)
        else:
            self.debug("Wrong instance calling wrong function on_numeric_change")

    def on_categorical_change(self):
        if isinstance(self.filter_widget, MultiSelectComboBox):
            self.filterstore.filter_value = self.filter_widget.currentData()
            self.debug(self.filterstore.filter_value)
        else:
            self.debug("Wrong instance calling wrong function on_categorical_change")

    def on_datetime_change(self):
        try:
            self.filterstore.filter_value = (
                pd.to_datetime(self.min_value.dateTime().toPython()),  # pyright: ignore[reportCallIssue, reportUnknownMemberType, reportArgumentType]
                pd.to_datetime(self.max_value.dateTime().toPython()),  # pyright: ignore[reportCallIssue, reportUnknownMemberType, reportArgumentType]
            )
            self.debug(self.filterstore.filter_value)
        except AttributeError:
            self.debug("Wrong instance calling wrong function on_datetime_change")

    def onDelete(self):
        self.datastore.on_filter_widget_delete(self)
        _ = self.close()



from dataclasses import dataclass
from enum import Enum

IMPORT_FILE_TYPES = [
    "Excel Files (*.xlsx *.xls *.xlsm)",
    "CSV (*.csv)",
    "Parquet (*.parquet)",
    "Feather (*.feather)",
]

IMPORT_FILE_TYPES_LIST = [".xlsx", ".xls", ".xlsm", ".csv", ".parquet", ".feather"]

EXPORT_FILE_TYPES = [
    "Excel File (*.xlsx)",
    "CSV (*.csv)",
    "Parquet (*.parquet)",
    "Feather (*.feather)",
]
VALID_DTYPES = [
    "Categorical",
    "Numeric",
    "Datetime",
]


NAN_NUM = [
    "Replace with 0",
    "Drop rows",
    "Keep empty",
]
NAN_CAT = [
    "Replace with no data",
    "Drop rows",
    "Keep empty",
]
NAN_DATETIME = [
    "Drop rows",
    "Keep empty",
]
NAN_OPS = list(dict.fromkeys(NAN_NUM + NAN_CAT + NAN_DATETIME))

NUM_TO_CAT_OPS = [
    "As category",
    "Into n equal bins",
    "Bin by bin edges",
]


class NumericConversion(Enum):
    AS_CATEGORY = NUM_TO_CAT_OPS[0]
    BINNED = NUM_TO_CAT_OPS[1]
    BIN_WIDTH = NUM_TO_CAT_OPS[2]


@dataclass
class NumericConverter:
    conversion: NumericConversion
    value: int | list[float] | None


class Dtype(Enum):
    CATEGORICAL = VALID_DTYPES[0]
    NUMERIC = VALID_DTYPES[1]
    DATETIME = VALID_DTYPES[2]


DtypeOperation = Dtype


class NaNOperation(Enum):
    REPLACE_WITH_0 = NAN_OPS[0]
    DROP_ROWS = NAN_OPS[1]
    KEEP_AS_NaN = NAN_OPS[2]
    REPLACE_WITH_NO_DATA = NAN_OPS[3]



# pyright: reportUnknownMemberType=false, reportMissingTypeStubs=false
from collections.abc import Sequence
from typing import Callable, final, override

from PySide6.QtCore import QEvent, Qt
from PySide6.QtGui import QColor, QFontMetrics, QPalette, QResizeEvent, QStandardItem
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLayout,
    QLineEdit,
    QRadioButton,
    QScrollArea,
    QSlider,
    QSpacerItem,
    QSpinBox,
    QStyledItemDelegate,
    QVBoxLayout,
    QWidget,
)
from qframelesswindow import StandardTitleBar
from superqt import QElidingLabel, QLabeledDoubleRangeSlider, QLabeledDoubleSlider

widget_types = QLayout | QSlider | QWidget | QFrame | QSpacerItem


def build_layout(
    layout: QHBoxLayout | QVBoxLayout,
    widget_list: Sequence[widget_types | tuple[widget_types, int]],
):
    for w in widget_list:
        if isinstance(w, tuple):
            w, stretch = w
            if isinstance(w, QLayout):
                layout.addLayout(w, stretch=stretch)
            elif isinstance(w, QWidget):
                layout.addWidget(w, stretch=stretch)
        elif isinstance(w, QSpacerItem):
            layout.addSpacerItem(w)
        elif isinstance(w, QLayout):
            layout.addLayout(w)
        else:
            layout.addWidget(w)


def build_grid_layout(
    layout: QVBoxLayout,
    widget_lists: Sequence[Sequence[widget_types | tuple[widget_types, int]]],
):
    hboxes: list[QHBoxLayout] = []
    for wl in widget_lists:
        hboxes.append(QHBoxLayout())
        build_layout(hboxes[-1], wl)

    build_layout(layout, hboxes)


def get_dynamic_scroll_area(
    widget: QWidget | Callable[[bool], QWidget], width: int | None = None
) -> tuple[QScrollArea, QWidget]:
    scroll_area = QScrollArea()
    scroll_area.setObjectName("ScrollArea")
    scroll_area.setWidgetResizable(True)
    if callable(widget):
        widget = widget(False)
    scroll_area.setWidget(widget)
    if width:
        height = widget.height()
        widget.resize(width, height)
    widget.setObjectName("ScrollAreaWidget")
    return (scroll_area, widget)


def get_label_widget_row(
    label: str,
    widget: QWidget,
    setStretch: bool = False,
    useEliding: bool = False,
) -> QHBoxLayout:
    if useEliding:
        qlabel = QElidingLabel(label)
    else:
        qlabel = QLabel(label)
    layout = QHBoxLayout()
    if setStretch:
        build_layout(layout, [(qlabel, 1), (widget, 1)])
    else:
        build_layout(layout, [qlabel, widget])
    return layout


def get_label_widget_row_callback(
    label: str,
    widget: QWidget,
    callback: Callable[[], None],
    setStretch: bool = False,
    useEliding: bool = False,
) -> QHBoxLayout:
    add_callback_to_standard_signal([widget], callback)
    return get_label_widget_row(label, widget, setStretch, useEliding)


def build_layout_with_callbacks(
    layout: QHBoxLayout | QVBoxLayout,
    widget_list: Sequence[widget_types | tuple[widget_types, int]]
    | Sequence[Sequence[widget_types | tuple[widget_types, int]]],
    callback: Callable[[], None],
):
    if all(isinstance(wl, list) for wl in widget_list):
        if not isinstance(layout, QVBoxLayout):
            raise ValueError("build_layout_with_callbacks got a wrong Layout Type")
        build_grid_layout(layout, widget_list)  # pyright: ignore[reportArgumentType]
        add_callback_to_standard_signal(widget_list, callback)
    else:
        build_layout(layout, widget_list)  # pyright: ignore[reportArgumentType]
        add_callback_to_standard_signal(widget_list, callback)


callback_types = (
    QComboBox
    | QLineEdit
    | QSpinBox
    | QRadioButton
    | QLabeledDoubleRangeSlider
    | QLabeledDoubleSlider
    | QDoubleSpinBox
    | QCheckBox
)


def add_callback_to_standard_signal(
    widget_list: Sequence[callback_types | widget_types | tuple[widget_types, int]]
    | Sequence[Sequence[callback_types | widget_types | tuple[widget_types, int]]],
    callback: Callable[[], None],
):
    for widget in widget_list:
        if isinstance(widget, tuple):
            widget, _ = widget
        match widget:
            case list():
                add_callback_to_standard_signal(widget, callback)
            case MultiSelectComboBox():
                _ = widget.dataChanged.connect(callback)
            case QComboBox():
                _ = widget.currentTextChanged.connect(callback)
            case QLineEdit():
                _ = widget.editingFinished.connect(callback)
            case QSpinBox():
                _ = widget.valueChanged.connect(callback)
            case QDoubleSpinBox():
                _ = widget.valueChanged.connect(callback)
            case QRadioButton():
                _ = widget.toggled.connect(callback)
            case QLabeledDoubleRangeSlider():
                _ = widget.valueChanged.connect(callback)
            case QLabeledDoubleSlider():
                _ = widget.valueChanged.connect(callback)
            case QCheckBox():
                _ = widget.toggled.connect(callback)
            case _:
                pass


class MultiSelectComboBox(QComboBox):
    """
    (Designed by user0706, repurposed for PySide6:
    https://github.com/user0706/pyqt6-multiselect-combobox)
    A custom combo box widget that allows for multi-select functionality.

    This widget provides the ability to select multiple items from a dropdown list
    and display them in a comma-separated format in the combo box's line edit area.

    MIT License (from original)

    Copyright (c) 2024 Marko

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
    """

    class Delegate(QStyledItemDelegate):
        def sizeHint(self, option, index):
            size = super().sizeHint(option, index)
            size.setHeight(20)
            return size

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.placeholderText = ""

        self.setEditable(True)
        self.lineEdit().setReadOnly(True)

        palette = self.lineEdit().palette()
        palette.setBrush(
            QPalette.ColorRole.Base, palette.brush(QPalette.ColorRole.Button)
        )
        self.lineEdit().setPalette(palette)

        self.setSizeAdjustPolicy(
            QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon
        )
        self.setItemDelegate(MultiSelectComboBox.Delegate())
        self.setOutputType("data")
        self.setDisplayType("data")
        self.setDisplayDelimiter(",")

        self.model().dataChanged.connect(self.updateText)
        self.lineEdit().installEventFilter(self)
        self.closeOnLineEditClick = False
        self.view().viewport().installEventFilter(self)

    @property
    def dataChanged(self):
        return self.model().dataChanged

    def setOutputType(self, output_type: str) -> None:
        """Set the output type for the combo box.

        Args:
            output_type (str): The output type, either 'data' or 'text'.

        Raises:
            ValueError: If output_type is neither 'data' nor 'text'.
        """
        if output_type in ["data", "text"]:
            self.output_type = output_type
        else:
            raise ValueError("Output type must be 'data' or 'text'")

    def setDisplayType(self, display_type: str) -> None:
        """Set the display type for the combo box.

        Args:
            display_type (str): The display type, either 'data' or 'text'.

        Raises:
            ValueError: If display_type is neither 'data' nor 'text'.
        """
        if display_type in ["data", "text"]:
            self.display_type = display_type
        else:
            raise ValueError("Display type must be 'data' or 'text'")

    def getOutputType(self) -> str:
        """Get the current output type.

        Returns:
            str: The current output type, either 'data' or 'text'.
        """
        return self.output_type

    def getDisplayType(self) -> str:
        """Get the current display type.

        Returns:
            str: The current display type, either 'data' or 'text'.
        """
        return self.display_type

    def setDisplayDelimiter(
        self, delimiter: str, space_after: bool = True, space_before: bool = False
    ) -> None:
        """Set the display delimiter for the combo box.

        Args:
            delimiter (str): The delimiter to use.
        """
        sufix = " " if space_after else ""
        prefix = " " if space_before else ""
        self.display_delimiter = prefix + delimiter + sufix

    def getDisplayDelimiter(self) -> str:
        """Get the current display delimiter.

        Returns:
            str: The current display delimiter.
        """
        return self.display_delimiter

    def resizeEvent(self, event) -> None:
        """Resize event handler.

        Args:
            event: The resize event.
        """
        self.updateText()
        super().resizeEvent(event)

    def eventFilter(self, obj, event) -> bool:
        """Event filter to handle mouse button release events.

        Args:
            obj: The object emitting the event.
            event: The event being emitted.

        Returns:
            bool: True if the event was handled, False otherwise.
        """
        if obj == self.lineEdit() and event.type() == QEvent.Type.MouseButtonRelease:
            if self.closeOnLineEditClick:
                self.hidePopup()
            else:
                self.showPopup()
            return True
        if (
            obj == self.view().viewport()
            and event.type() == QEvent.Type.MouseButtonRelease
        ):
            index = self.view().indexAt(event.position().toPoint())
            item = self.model().itemFromIndex(index)
            if item.checkState() == Qt.CheckState.Checked:
                item.setCheckState(Qt.CheckState.Unchecked)
            else:
                item.setCheckState(Qt.CheckState.Checked)
            return True
        return False

    def showPopup(self) -> None:
        """Show the combo box popup."""
        super().showPopup()
        self.closeOnLineEditClick = True

    def hidePopup(self) -> None:
        """Hide the combo box popup."""
        super().hidePopup()
        self.startTimer(100)

    def timerEvent(self, event) -> None:
        """Timer event handler.

        Args:
            event: The timer event.
        """
        self.killTimer(event.timerId())
        self.closeOnLineEditClick = False

    def typeSelection(
        self, index: int, type_variable: str, expected_type: str = "data"
    ) -> str:
        """Get the selected item's data or text based on type.

        Args:
            index (int): The index of the item.
            type_variable (str): The type variable to retrieve ('data' or 'text').
            expected_type (str): The expected type. Default is 'data'.

        Returns:
            str: The selected item's data or text.
        """
        if type_variable == expected_type:
            return self.model().item(index).data()
        return self.model().item(index).text()

    def updateText(self) -> None:
        """Update the displayed text based on selected items."""
        display_type = self.getDisplayType()
        delimiter = self.getDisplayDelimiter()
        texts = [
            self.typeSelection(i, display_type)
            for i in range(self.model().rowCount())
            if self.model().item(i).checkState() == Qt.CheckState.Checked
        ]

        if texts:
            text = delimiter.join(texts)
        else:
            text = self.placeholderText if hasattr(self, "placeholderText") else ""

        metrics = QFontMetrics(self.lineEdit().font())
        elidedText = metrics.elidedText(
            text, Qt.TextElideMode.ElideRight, self.lineEdit().width()
        )
        self.lineEdit().setText(elidedText)

    def addItem(self, text: str, data: str = None) -> None:
        """Add an item to the combo box.

        Args:
            text (str): The text to display.
            data (str): The associated data. Default is None.
        """
        item = QStandardItem()
        item.setText(text)
        item.setData(data or text)
        item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable)
        item.setData(Qt.CheckState.Unchecked, Qt.ItemDataRole.CheckStateRole)
        self.model().appendRow(item)

    def addItems(self, texts: list, dataList: list = None) -> None:
        """Add multiple items to the combo box.

        Args:
            texts (list): A list of texts to display.
            dataList (list): A list of associated data. Default is None.
        """
        dataList = dataList or [None] * len(texts)
        for text, data in zip(texts, dataList):
            self.addItem(text, data)

    def currentData(self) -> list:
        """Get the currently selected data.

        Returns:
            list: A list of currently selected data.
        """
        output_type = self.getOutputType()
        return [
            self.typeSelection(i, output_type)
            for i in range(self.model().rowCount())
            if self.model().item(i).checkState() == Qt.CheckState.Checked
        ]

    def setCurrentIndexes(self, indexes: list) -> None:
        """Set the selected items based on the provided indexes.

        Args:
            indexes (list): A list of indexes to select.
        """
        for i in range(self.model().rowCount()):
            self.model().item(i).setCheckState(
                Qt.CheckState.Checked if i in indexes else Qt.CheckState.Unchecked
            )
        self.updateText()

    def getCurrentIndexes(self) -> list:
        """Get the indexes of the currently selected items.

        Returns:
            list: A list of indexes of selected items.
        """
        return [
            i
            for i in range(self.model().rowCount())
            if self.model().item(i).checkState() == Qt.CheckState.Checked
        ]

    def setPlaceholderText(self, text: str) -> None:
        """Set the placeholder text for the combo box.

        Args:
            text (str): The placeholder text.
        """
        self.placeholderText = text
        self.updateText()

    def showEvent(self, event) -> None:
        """Show event handler.

        Args:
            event: The show event.
        """
        super().showEvent(event)
        self.updateText()

    def getCurrentOptions(self):
        """Get the currently selected options along with their associated data.

        Returns:
            list: A list of tuples containing the text and data
            of the currently selected options.
                Each tuple consists of (text, data).
        """
        res = []
        for i in range(self.model().rowCount()):
            if self.model().item(i).checkState() == Qt.CheckState.Checked:
                res.append((self.model().item(i).text(), self.model().item(i).data()))
        return res

    def getPlaceholderText(self):
        """Get the placeholder text currently set for the combo box.

        Returns:
            str: The placeholder text.
        """
        return self.placeholderText

    def setDuplicatesEnabled(self, enabled: bool) -> None:
        """Set whether duplicates are allowed in the combo box.

        Args:
            enabled (bool): Whether duplicates are allowed.
        """
        self.duplicatesEnabled = enabled

    def isDuplicatesEnabled(self) -> bool:
        """Check if duplicates are allowed in the combo box.

        Returns:
            bool: True if duplicates are allowed, False otherwise.
        """
        return self.duplicatesEnabled


@final
class CustomTitleBar(StandardTitleBar):
    def __init__(self, parent: QWidget):
        super().__init__(parent)

        self.text = QLabel("")
        layout = self.layout()
        if layout is not None:
            self.text.setStyleSheet("""
                QLabel {
                    font-size: 14px;
                    color: #FEEEE8;
                    background-color: transparent
                }
            """)
            self.text.setAlignment(Qt.AlignmentFlag.AlignCenter)
            layout.insertWidget(2, self.text, 0, Qt.AlignmentFlag.AlignCenter)  # pyright: ignore[reportAttributeAccessIssue]
        self.minBtn.setNormalColor(QColor("gray"))
        self.maxBtn.setNormalColor(QColor("gray"))
        self.closeBtn.setNormalColor(QColor("gray"))

    @override
    def resizeEvent(self, event: QResizeEvent, /) -> None:
        self.text.move(
            (self.width() - self.text.width()) // 2,
            (self.height() - self.text.height()) // 2,
        )
        return super().resizeEvent(event)

    def changeTitle(self, text: str):
        self.text.setText(text)
        oldw = self.width()
        oldh = self.height()
        self.resize(self.width() - 1, self.height() - 1)
        self.resize(oldw, oldh)



# pyright: reportArgumentType=false, reportUnknownMemberType=false, reportMissingTypeStubs=false
from pathlib import Path

import pandas as pd
from spoofs import debug_error_spoof as debug_spoof

from data_explorer_qt.data.dataenums import Dtype
from data_explorer_qt.data.datamodel import (
    FilterStore,
    apply_filter,
    categorical_comparator,
    datetime_comparator,
    handle_dtype_operation,
    handle_nan_operation,
    numeric_comparator,
)


def test_dtype_conversion():
    # 6 possible conversions, but only 3 valid conversions
    # Any -> Categorical, Numeric -> Categorical, Categorical -> Datetime.
    path_to_data = str(Path(__file__).parent / "Test.xlsx")
    data = pd.read_excel(path_to_data, engine="calamine")
    # Dates column is default read as categorical in this case.
    assert not datetime_comparator(data["Dates"])
    # Categorical -> Datetime
    assert datetime_comparator(
        handle_dtype_operation(data["Dates"], "Datetime", debug_spoof)
    )
    # Numeric -> Categorical
    assert categorical_comparator(
        handle_dtype_operation(data["Header 1"], "Categorical", debug_spoof)
    )
    # Categorical (empty column) -> Numeric
    assert numeric_comparator(
        handle_dtype_operation(data["Header 2"], "Numeric", debug_spoof)
    )


def test_nan_handling():
    path_to_data = str(Path(__file__).parent / "TestNaN.xlsx")
    data = pd.read_excel(path_to_data, engine="calamine")
    data["Dates"] = handle_dtype_operation(data["Dates"], "Datetime", debug_spoof)
    data = handle_nan_operation(data, "Keep as NaN", "Dates", debug_spoof)
    assert data["Dates"].isna().sum() == 2
    data = handle_nan_operation(data, "Replace with No Data", "Header 3", debug_spoof)
    assert data["Header 3"].str.contains("No Data").sum() == 3
    data = handle_nan_operation(data, "Replace with 0", "Header 2", debug_spoof)
    assert (data["Header 2"] == 0).sum() == len(data)


def test_filtering():
    path_to_data = str(Path(__file__).parent / "Test.xlsx")
    filter = FilterStore(dtype=Dtype.CATEGORICAL, filter_value=["Hello"])
    data = pd.read_excel(path_to_data, engine="calamine")
    data = apply_filter(data, column="Header 3", filterstore=filter)
    assert len(data) == 4
    data = pd.read_excel(path_to_data, engine="calamine")
    filter = FilterStore(dtype=Dtype.NUMERIC, filter_value=(10, 100))
    data = apply_filter(data, column="Header 1", filterstore=filter)
    assert len(data) == 2



# pyright: reportImportCycles=false, reportUnusedImport=false
import importlib
import logging
import os
import pprint
import sys
import traceback
from pathlib import Path
from typing import Any, Callable, final

from PySide6.QtCore import Qt
from PySide6.QtGui import QFontDatabase, QIcon
from PySide6.QtWidgets import QApplication, QWidget
from qframelesswindow import FramelessWindow

from data_explorer_qt.guihelper import CustomTitleBar

from . import setqtapi  # noqa
from .config import CONFIG
from .data.datamodel import DataModel
from .dataexplorerGUI import DataExplorerGUI
from .plotter.plotter import Plotter


@final
class DataExplorer:
    stylesheet: str = ""
    config = CONFIG
    owned_widgets: list[QWidget] = []
    plugin_list: list[Callable[["DataExplorer"], tuple[QWidget, str]]] = []

    def __init__(self, app: QApplication, debug_mode: bool):
        self.app = app
        self.debug_mode = debug_mode
        self.font = self.config["General"]["font"]
        self.font_size = self.config["General"]["fontsize"]
        self._setup_logging()
        self.icon_path = str(Path(__file__).parent / "resources" / "icon.ico")

        self.construct_plugin_list()

        self.load_custom_font()

        self.gui = DataExplorerGUI(self)
        self.info("GUI Initialised!")

        self.datamodel = DataModel(self)
        self.info("Data Model Initialised!")

        self.plotter = Plotter(self)
        self.info("Plotter Initialised")

    def _setup_logging(self):
        # Parent of this file is the upper directory
        # Parent of that is the upper directory
        log_folder = Path.home() / ".data-explorer-qt" / "Logs"

        log_folder.mkdir(parents=True, exist_ok=True)

        log_file_number = 1
        log_file_path = log_folder / "Log.log"
        while log_file_path.exists():
            log_file_path = log_folder / f"Log {log_file_number}.log"
            log_file_number += 1

        self._logger = logging.getLogger("DataExplorer")
        self._logger.propagate = False
        self._logger.setLevel(logging.DEBUG)

        log_format = logging.Formatter(self.config["Logger"]["log_format"])
        console = logging.StreamHandler(sys.stdout)
        console.setFormatter(log_format)
        file = logging.FileHandler(log_file_path)
        file.setFormatter(log_format)
        file.setLevel(logging.DEBUG)
        if self.debug_mode:
            console.setLevel(logging.DEBUG)
        else:
            console.setLevel(logging.INFO)
        self._logger.addHandler(console)
        self._logger.addHandler(file)
        self._logger.info("Logger Setup!")

    def debug(self, msg_or_object: str | Any):
        if isinstance(msg_or_object, str):
            self._logger.debug(msg_or_object)
        else:
            self._logger.debug(pprint.pformat(msg_or_object))

    def error(self, msg_or_object: str | Any):
        if isinstance(msg_or_object, str):
            self.gui.error_msg(msg_or_object)
            self._logger.error(msg_or_object)
        else:
            self._logger.error(pprint.pformat(msg_or_object))
            self.gui.error_msg(pprint.pformat(msg_or_object))

    def info(self, msg_or_object: str | Any):
        if isinstance(msg_or_object, str):
            self._logger.info(msg_or_object)
        else:
            self._logger.info(pprint.pformat(msg_or_object))

    def status_message(self, msg: str):
        self._logger.info(msg)
        self.gui.write_to_status_bar(msg)

    def construct_plugin_list(self):
        plugin_path: str = self.config["General"]["plugins"]
        if not Path(plugin_path).exists():
            self.debug("Plugin Path is invalid.")
            plugin_path_ = Path(__file__).parent.parent / "plugins"
            if not plugin_path_.exists():
                return
            else:
                self.info("Using default plugin path.")
                plugin_path = str(plugin_path_)
        self.debug(plugin_path)
        sys.path.insert(0, plugin_path)
        for filename in os.listdir(plugin_path):
            if filename.endswith(".py") and filename != "__init__.py":
                module_name = filename.split(".")[0]
                module_name = f"{module_name}"
                try:
                    module = importlib.import_module(module_name)
                    self.plugin_list.append(module.init_plugin)
                    self.info(f"Loaded {module_name}")
                except Exception as e:
                    self.debug(f"Plugin {module_name} not loaded")
                    self.debug(e)

    def load_custom_font(self):
        path_to_font = Path(__file__).parent / "resources" / "PTRootUI.ttf"
        if path_to_font.exists():
            path_to_font = str(path_to_font)
            font_id = QFontDatabase.addApplicationFont(path_to_font)
            if font_id != -1:
                self.debug(QFontDatabase.applicationFontFamilies(font_id)[0])
                self.debug("Loaded Custom Font")
        else:
            return

    def closeEvent(self):
        self.app.quit()

    def get_widget(self, detached: bool = False) -> QWidget:
        if detached:
            widget = FramelessWindow()
            cust_title = CustomTitleBar(widget)
            widget.setTitleBar(cust_title)
        else:
            widget = QWidget()
        widget.setWindowIcon(QIcon(self.icon_path))
        widget.setObjectName("StandardWidget")
        widget.setStyleSheet(self.stylesheet)
        return widget


def run():
    try:
        app = QApplication([])
        app.setAttribute(Qt.ApplicationAttribute.AA_DontCreateNativeWidgetSiblings)
        argv = sys.argv
        _ = app.setStyle("Fusion")
        debug_mode = False
        if argv is None:
            argv = []
        if "--debug" in argv:
            debug_mode = True
        dataexplorer = DataExplorer(app, debug_mode)
        dataexplorer.gui.show()
        sys.exit(app.exec())
    except Exception:
        traceback.print_exc()
        sys.exit(1)



from pathlib import Path
from typing import Any

import toml

starting_path = Path(__file__)
config_path = starting_path.parent / "config.toml"


def get_config() -> dict[str, Any]:
    config = toml.load(config_path)
    config = construct_stylesheets(config)
    return config


def construct_stylesheets(config: dict[str, Any]) -> dict[str, Any]:
    for key in config["Themes"]:
        assert isinstance(key, str)
        config["Themes"][key] = _theme_constructor(config["Themes"][key])
    return config


def _theme_constructor(theme_input: dict[str, str] | str) -> str:
    if isinstance(theme_input, dict):
        return "\n".join([_theme_constructor(theme_input[key]) for key in theme_input])
    else:
        return theme_input


def get_list_of_themes(config: dict[str, Any]):
    return list(config["Themes"].keys())


CONFIG = get_config()
CONFIG_STYLESHEET = CONFIG["Themes"]["Light"]



# pyright: reportUnknownMemberType=false
import pprint
import traceback
import typing
from dataclasses import dataclass
from itertools import product
from typing import Callable, Literal, override

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.backends.backend_qt import NavigationToolbar2QT
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from PySide6.QtGui import QCloseEvent, QIcon
from PySide6.QtWidgets import QComboBox, QHBoxLayout, QTextEdit, QVBoxLayout, QWidget
from qframelesswindow import FramelessWindow
from seaborn import FacetGrid, PairGrid

from ..data.dataenums import NumericConversion
from ..guihelper import CustomTitleBar, MultiSelectComboBox, build_layout

if typing.TYPE_CHECKING:
    from ..data.datamodel import DataStore
    from ..dataexplorer import DataExplorer

FILTER_DISPLAY_STARTING_TEXT = "Filters: "

MARKERS = [
    " ",
    ".",
    ",",
    "o",
    "v",
    "^",
    "<",
    ">",
    "1",
    "2",
    "3",
    "4",
    "8",
    "s",
    "p",
    "P",
    "*",
    "h",
    "H",
    "+",
    "x",
    "X",
    "D",
    "d",
]

LINE_STYLES = ["-", "--", "-.", ":", "None"]

VIOLIN_INNER = ["box", "quart", "point", "stick"]

HIST_PLOT_STATISTICS = ["count", "frequency", "proportion", "percent", "density"]

HIST_MULTIPLE = ["layer", "dodge", "stack", "fill"]

COUNT_PLOT_STATISTICS = ["count", "proportion", "percent"]

CORREL_STATISTICS = ["pearson", "kendall", "spearman"]

COLOR_PALETTES = {
    "qualitative": [
        "tab10",
        "deep",
        "muted",
        "pastel",
        "bright",
        "dark",
        "colorblind",
        "tab20",
        "tab20b",
        "tab20c",
    ],
    "circular": ["hls", "husl"],
    "perceptually_uniform": [
        "rocket",
        "mako",
        "flare",
        "crest",
        "viridis",
        "plasma",
        "inferno",
        "magma",
        "cividis",
    ],
    "diverging": ["vlag", "icefire", "coolwarm", "bwr", "seismic"],
}

PALETTE_TYPES = ["qualitative", "circular", "perceptually_uniform", "diverging"]

SORT_CATEGORIES = [
    "First Occurence",
    "Alphabetical",
]

REGRESSION_MIN_DEGREE = 1
REGRESSION_MAX_DEGREE = 4


@typing.final
@dataclass
class TickParams:
    axis: Literal["x", "y"]
    rotation: int
    grid_colour: str
    grid_alpha: float

    def to_kwargs(self) -> dict[str, int | str | float]:
        return {
            "axis": self.axis,
            "labelrotation": self.rotation,
            "grid_color": self.grid_colour,
            "grid_alpha": self.grid_alpha,
        }


@typing.final
class EmbeddedDynamicPlot(QWidget):
    filter_display: QTextEdit
    plot: QWidget
    figure: Figure

    def __init__(
        self,
        dataexplorer: "DataExplorer",
        datastore: "DataStore",
        name: str,
        parent: "PlottingDialog",
    ):
        super().__init__()
        self.setWindowTitle(name)
        self.debug = dataexplorer.debug
        self.error = dataexplorer.error
        self.dataexplorer = dataexplorer
        self._parent = parent
        self.name = name
        self.resize(1200, 1200)
        self.setObjectName("DynamicPlot")
        self.setStyleSheet(dataexplorer.stylesheet)
        self.datastore = datastore
        self.filter_display = QTextEdit()
        self.filter_display.setReadOnly(True)
        self.filter_display.setText(self._generate_filter_text())
        self.dataexplorer.owned_widgets.append(self)
        self.setWindowIcon(QIcon(self.dataexplorer.icon_path))

        self.plot_subwidget = self.dataexplorer.get_widget()
        self.figure = plt.figure()
        self.plot = FigureCanvas(plt.figure())

        self.plot_toolbar: QWidget = NavigationToolbar2QT(self.plot)

        self.plot_vbox = QVBoxLayout(self.plot_subwidget)

        build_layout(self.plot_vbox, [self.plot_toolbar, self.plot])

        self._layout = QHBoxLayout(self)
        build_layout(self._layout, [(self.plot_subwidget, 1), self.filter_display])

        self.show()

    def update_dynamic_widget(self, plot: Figure | FacetGrid):
        self._draw_plot(plot)
        self.filter_display.setText(self._generate_filter_text())

    def _draw_plot(self, plot: Figure | FacetGrid):
        old_plot = self.plot
        old_toolbar = self.plot_toolbar
        old_figure = self.figure
        old_plot_subwidget = self.plot_subwidget
        if isinstance(plot, Figure):
            self.figure = plot
        else:
            self.figure = plot.figure

        plt.close(old_figure)
        old_plot.close()
        old_toolbar.close()
        self.plot = FigureCanvas(self.figure)
        self.plot_toolbar = NavigationToolbar2QT(self.plot)

        self.plot_subwidget = self.dataexplorer.get_widget()
        self.plot_vbox = QVBoxLayout(self.plot_subwidget)
        build_layout(self.plot_vbox, [self.plot_toolbar, self.plot])
        _ = self._layout.replaceWidget(old_plot_subwidget, self.plot_subwidget)
        old_plot_subwidget.close()
        self.update()
        self.debug(f"{self.name} updated!")

    def _generate_filter_text(self) -> str:
        filter_text = FILTER_DISPLAY_STARTING_TEXT
        for column in self.datastore.filters:
            filter_text += f"<br>- {column}"  # Newline + bullet list start
            for fs in self.datastore.filters[column]:
                if fs.active:
                    filter_text += pprint.pformat(fs.filter_value)
        return filter_text

    @typing.override
    def closeEvent(self, event: QCloseEvent):
        self.dataexplorer.owned_widgets.remove(self)
        self._parent.delete_dynamic_plot()
        return super().closeEvent(event)


class PlottingDialog(FramelessWindow):
    dynamic_plot_widget: EmbeddedDynamicPlot | None = None
    dynamic_callback_id: int = -1
    plotting_data: pd.DataFrame
    palette_type: Literal[*PALETTE_TYPES] = PALETTE_TYPES[0]  # pyright: ignore

    def __init__(self, dataexplorer: "DataExplorer", datastore: "DataStore", name: str):
        super().__init__()
        self.setWindowTitle(f"{name} Plotting Dialog")
        cust_title = CustomTitleBar(self)
        self.setTitleBar(cust_title)
        self.setObjectName("StandardWidget")
        cust_title.changeTitle(f"{name} Plotting Dialog")
        self.dataexplorer: "DataExplorer" = dataexplorer
        self.debug: Callable[[str], None] = dataexplorer.debug
        self.error: Callable[[str], None] = dataexplorer.error
        self.datastore: "DataStore" = datastore
        self.setWindowIcon(QIcon(self.dataexplorer.icon_path))

        self._generate_plotting_data()
        self.dataexplorer.owned_widgets.append(self)
        self._layout: QVBoxLayout = QVBoxLayout(self)
        self.setStyleSheet(dataexplorer.stylesheet)
        self.resize(600, 800)

    def setup_column_combobox(self, mandatory: bool, multi_variable: bool = False):
        if multi_variable:
            box = MultiSelectComboBox()
        else:
            box = QComboBox()
            box.setSizeAdjustPolicy(
                QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon
            )
        if mandatory:
            box.addItems(self.datastore.columns)
        else:
            box.addItem("")
            box.addItems(self.datastore.columns)
        return box

    def cat_name(self, col_name: str):
        return col_name + "_categorical_"

    def _generate_plotting_data(self):
        self.plotting_data = self.datastore.filtered_data.copy()
        cat_name = self.cat_name

        for column, nc in self.datastore.numeric_to_categorical.items():
            self.debug(f"{column} {nc.conversion} {nc.value}")
            match nc.conversion:
                case NumericConversion.AS_CATEGORY:
                    self.plotting_data[cat_name(column)] = self.plotting_data[column]
                case NumericConversion.BINNED:
                    if not isinstance(nc.value, int):
                        self.error("Incorrect type for numeric->categorical bin_N")
                        return
                    self.plotting_data[cat_name(column)] = pd.cut(
                        self.plotting_data[column], nc.value
                    )
                case NumericConversion.BIN_WIDTH:
                    if not isinstance(nc.value, list):
                        self.error("Incorrect type for numeric->categorical bin edges")
                        return
                    self.plotting_data[cat_name(column)] = pd.cut(
                        self.plotting_data[column], nc.value
                    )

    def get_categorical_column_name(self, col_name: str):
        if col_name == "":
            return None
        else:
            cat_name = self.cat_name(col_name)
            if cat_name in self.plotting_data.columns:
                return cat_name
            else:
                return col_name

    def get_category_order(self, col_name: str | None):
        if col_name is None or col_name.endswith("_categorical_"):
            return None
        match self.dataexplorer.plotter.sort_category_by:
            case "First Occurence":
                return None
            case "Alphabetical":
                return sorted(self.plotting_data[col_name].unique())

    def get_palette(self):
        match self.palette_type:
            case "qualitative":
                return self.dataexplorer.plotter.qualitative_palette
            case "circular":
                return self.dataexplorer.plotter.circular_palette
            case "perceptually_uniform":
                return self.dataexplorer.plotter.perceptually_uniform_palette
            case "diverging":
                return self.dataexplorer.plotter.diverging_palette

    def on_plot(self):
        self._generate_plotting_data()

    def plotter(self) -> Figure | FacetGrid | PairGrid: ...

    def plot(self):
        self.on_plot()
        try:
            fig_or_fg = self.plotter()
            if isinstance(fig_or_fg, Figure):
                fig_or_fg.show()
            else:
                fig_or_fg.figure.show()
        except TimeoutError:
            self.error("Plot took too long. Check your variables")
        except Exception:
            self.error("Unable to plot. Err: " + traceback.format_exc())

    def dynamic_plot(self): ...

    def redraw_dynamic_plot(self): ...

    def delete_dynamic_plot(self):
        self.debug("Dynamic Plot Widget Deleted")
        self.dynamic_plot_widget = None
        self.datastore.remove_filter_change_callback(self.dynamic_callback_id)
        self.dynamic_callback_id = -1

    def on_widget_change(self):
        if self.dynamic_plot_widget is not None:
            self.redraw_dynamic_plot()

    @override
    def closeEvent(self, event: QCloseEvent) -> None:
        self.dataexplorer.owned_widgets.remove(self)
        return super().closeEvent(event)


def get_dataframe_X_for_degree(
    df: pd.DataFrame, degree: int
) -> tuple[pd.DataFrame, list[str]]:
    generated_columns = []
    columns = list(df.columns)
    for deg in range(2, degree + 1):  # 2, 3, 4, ...
        cols_this_iter = []
        new_cols_this_iter = []
        for col in columns:
            new_col_name = f"{col}^{deg}"
            assert new_col_name not in columns
            cols_this_iter.append(col)
            new_cols_this_iter.append(new_col_name)
        df[new_cols_this_iter] = df[cols_this_iter] ** deg
        generated_columns.extend(new_cols_this_iter)
    return (df, generated_columns)


def get_dataframe_X_with_interaction(
    df: pd.DataFrame, generated_columns: list[str]
) -> pd.DataFrame:
    columns = set(df.columns) - set(generated_columns)
    generated_columns = []
    for col_A, col_B in product(columns, columns):
        if col_A != col_B:
            new_col_name = f"{col_A}*{col_B}"
            if new_col_name in generated_columns:
                continue
            assert new_col_name not in columns
            df[new_col_name] = df[col_A] * df[col_B]
            generated_columns.extend([new_col_name, f"{col_B}*{col_A}"])

    return df



# pyright: reportMissingTypeStubs=false, reportUnknownVariableType=false, reportArgumentType=false, reportAssignmentType=false, reportOperatorIssue=false
import typing
from dataclasses import dataclass
from functools import cache, cached_property
from typing import Callable, final

import pandas as pd
from pandas.api.types import is_datetime64_any_dtype, is_numeric_dtype, is_object_dtype
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QComboBox,
    QGridLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

if typing.TYPE_CHECKING:
    from ..dataexplorer import DataExplorer

from ..guihelper import build_layout, get_dynamic_scroll_area
from .dataenums import Dtype, DtypeOperation, NaNOperation, NumericConverter
from .filterwidget import FilterStore, FilterWidget, InvalidFilterWidgetError
from .numcatwidget import NumCatWidget


@dataclass
class FilterGUI:
    filter_page: QWidget
    filter_layout: QGridLayout
    shape_label: QLabel
    filter_widgets: list[FilterWidget]


@dataclass
class NumCatGUI:
    num_to_cat_page: QWidget
    num_to_cat_layout: QGridLayout
    num_to_cat_widgets: list[NumCatWidget]


@final
class DataStore:
    name: str
    # Maps type of cleaning (ex: Nan) to dict of column names
    # to operation performed.
    cleaning_operations: dict[str, dict[str, str]]
    cleaned_data: pd.DataFrame
    # Maps column names to a filter operation
    # For datetimes and continuous variables,
    # this is a less than, greater than comparison.
    # For categorical variables,
    # this is a value in/value equal to comparison.
    filters: dict[str, list[FilterStore]]
    filterGUI: FilterGUI
    on_filter_change_callbacks: dict[int, Callable[[], None]]
    callback_id: int
    filtered_data: pd.DataFrame
    numeric_to_categorical: dict[str, NumericConverter]

    def __init__(
        self,
        cl_ops: dict[str, dict[str, str]],
        data: pd.DataFrame,
        name: str,
        dataexplorer: "DataExplorer",
    ):
        self.name = name
        self.cleaning_operations = cl_ops
        self.cleaned_data = data
        self.dataexplorer = dataexplorer
        self.debug = dataexplorer.debug
        self.error = dataexplorer.error
        self.filters = {}
        self.filterGUI = self._get_filter_GUI(name)
        self.numCatGUI = self._get_num_cat_GUI(name)
        self.filtered_data = data.copy()
        self.callback_id = 1
        self.on_filter_change_callbacks = {}
        self.numeric_to_categorical = {}

    def apply_filters(self):
        init_len = len(self.filtered_data)
        self.filtered_data = self.cleaned_data.copy()
        for column in self.filters:
            for filterstore in self.filters[column]:
                if not filterstore.active:
                    continue
                starting_length = len(self.filtered_data)
                self.debug(filterstore)
                self.filtered_data = apply_filter(
                    self.filtered_data, column, filterstore
                )
                self.debug(
                    f"{starting_length - len(self.filtered_data)} "
                    "rows removed by filter "
                    f"{column} {filterstore.dtype} {filterstore.filter_value}"
                )

        delta = init_len - len(self.filtered_data)
        if delta > 0:
            self.dataexplorer.status_message(
                f"{delta} rows removed by newly applied filters"
            )
        elif delta == 0:
            self.dataexplorer.status_message(
                "Newly applied filters have not changed the data."
            )
        else:
            self.dataexplorer.status_message(
                f"{-delta} rows added by newly applied filters"
            )

        self.filterGUI.shape_label.setText(
            f"Filtered - Rows: {self.filtered_data.shape[0]},"
            f"Columns: {self.filtered_data.shape[1]}"
        )

        self.replot_callbacks()

    def add_filter_change_callback(self, callback: Callable[[], None]) -> int:
        callback_id = self.callback_id
        self.on_filter_change_callbacks[callback_id] = callback
        self.callback_id += 1
        return callback_id

    def remove_filter_change_callback(self, callback_id: int):
        _ = self.on_filter_change_callbacks.pop(callback_id)

    def _get_filter_GUI(self, name: str) -> FilterGUI:
        gui = FilterGUI(
            filter_page=self.dataexplorer.get_widget(),
            filter_layout=QGridLayout(),
            shape_label=QLabel(),
            filter_widgets=[],
        )
        vbox_layout = QVBoxLayout(gui.filter_page)
        vbox_layout.setSpacing(10)

        name_label = QLabel(name)
        name_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        unfiltered_shape_label = QLabel(
            f"Unfiltered - Rows: {self.cleaned_data.shape[0]}, "
            f"Columns: {self.cleaned_data.shape[1]}"
        )
        unfiltered_shape_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        gui.shape_label = QLabel(
            f"Filtered - Rows: {self.cleaned_data.shape[0]}, "
            f"Columns: {self.cleaned_data.shape[1]}"
        )
        gui.shape_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        add_widget_button = QPushButton("Add Filter...")
        _ = add_widget_button.clicked.connect(self._filter_widget_dialog)

        apply_filter_button = QPushButton("Apply Filters")
        _ = apply_filter_button.clicked.connect(self.apply_filters)

        scroll_area, content_widget = get_dynamic_scroll_area(
            self.dataexplorer.get_widget
        )

        content_widget.setLayout(gui.filter_layout)

        build_layout(vbox_layout, [name_label, unfiltered_shape_label, gui.shape_label])
        vbox_layout.addWidget(scroll_area, 1)
        build_layout(vbox_layout, [add_widget_button, apply_filter_button])

        gui.filter_page.setLayout(vbox_layout)
        return gui

    def _filter_widget_dialog(self):
        self._select_column = self.dataexplorer.get_widget(detached=True)
        self._select_column.setWindowTitle("Select column to filter")
        layout = QVBoxLayout()
        self._select_column.resize(500, 300)
        label = QLabel("Choose a column")
        self._column_chosen = QComboBox()
        self._column_chosen.addItems(self.columns)
        ok_button = QPushButton("Ok")
        _ = ok_button.clicked.connect(self._add_filter_widget)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        widg_list = [label, self._column_chosen, ok_button]
        build_layout(layout, widg_list)
        self._select_column.setLayout(layout)
        self._select_column.show()

    def _add_filter_widget(self):
        column = self._column_chosen.currentText()
        self._select_column.deleteLater()
        self.debug(f"Trying to create filter widget for {column}")
        try:
            filter_widget = FilterWidget(self, column, self.dataexplorer)
        except InvalidFilterWidgetError:
            return
        n = len(self.filterGUI.filter_widgets)
        self.filterGUI.filter_layout.addWidget(
            filter_widget, self._row(n), self._col(n)
        )
        if column in self.filters:
            self.filters[column].append(filter_widget.filterstore)
        else:
            self.filters[column] = [filter_widget.filterstore]
        self.filterGUI.filter_widgets.append(filter_widget)

    def _row(self, n: int) -> int:
        grid_max_col: int = self.dataexplorer.config["Grid"]["grid_max_col"]
        return int(n / grid_max_col)

    def _col(self, n: int) -> int:
        grid_max_col: int = self.dataexplorer.config["Grid"]["grid_max_col"]
        return n % grid_max_col

    def on_filter_widget_delete(self, widget: FilterWidget):
        layout = self.filterGUI.filter_layout
        ws = self.filterGUI.filter_widgets

        if len(ws) > 1 and widget != ws[-1]:
            for widg in ws:
                layout.removeWidget(widg)
            ws.remove(widget)
            self.debug("Widget Removed!")
            for i, widg in enumerate(ws):
                layout.addWidget(widg, self._row(i), self._col(i))
            layout.update()
            self.debug("Layout Updated!")
        else:
            layout.removeWidget(widget)
            self.debug("Last Widget Removed!")
            ws.remove(widget)

        self.filters[widget.column].remove(widget.filterstore)
        if len(self.filters[widget.column]) == 0:
            _ = self.filters.pop(widget.column)
        self.apply_filters()

    def _get_num_cat_GUI(self, name: str) -> NumCatGUI:
        gui = NumCatGUI(
            num_to_cat_page=self.dataexplorer.get_widget(),
            num_to_cat_layout=QGridLayout(),
            num_to_cat_widgets=[],
        )
        vbox_layout = QVBoxLayout(gui.num_to_cat_page)
        vbox_layout.setSpacing(10)

        name_label = QLabel(name)
        name_label.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        add_widget_button = QPushButton("Add numeric to categorical rule...")
        _ = add_widget_button.clicked.connect(self._num_cat_dialog)

        scroll_area, content_widget = get_dynamic_scroll_area(
            self.dataexplorer.get_widget
        )

        content_widget.setLayout(gui.num_to_cat_layout)

        build_layout(vbox_layout, [name_label])
        vbox_layout.addWidget(scroll_area, 1)
        build_layout(vbox_layout, [add_widget_button])

        gui.num_to_cat_page.setLayout(vbox_layout)
        return gui

    def _num_cat_dialog(self):
        self._select_column_nc = self.dataexplorer.get_widget(detached=True)
        self._select_column_nc.setWindowTitle("Select numeric column")
        layout = QVBoxLayout()
        self._select_column_nc.resize(200, 200)
        label = QLabel("Choose a column")
        self._column_chosen_nc = QComboBox()
        self.debug(self.numeric_columns)
        self._column_chosen_nc.addItems(self.numeric_columns)
        ok_button = QPushButton("Ok")
        _ = ok_button.clicked.connect(self._add_num_cat_widget)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        widg_list = [label, self._column_chosen_nc, ok_button]
        build_layout(layout, widg_list)
        self._select_column_nc.setLayout(layout)
        self._select_column_nc.show()

    def _add_num_cat_widget(self):
        column = self._column_chosen_nc.currentText()
        self._select_column_nc.deleteLater()
        self.debug(f"Trying to create num_cat widget for {column}")
        if column in self.numeric_to_categorical:
            self.error(
                "Cannot convert a numerical column to a categorical column twice"
            )
            return
        num_cat_widget = NumCatWidget(self, column, self.dataexplorer)
        n = len(self.numCatGUI.num_to_cat_widgets)
        self.numCatGUI.num_to_cat_layout.addWidget(
            num_cat_widget, self._row(n), self._col(n)
        )
        self.numeric_to_categorical[column] = num_cat_widget.numeric_converter
        self.numCatGUI.num_to_cat_widgets.append(num_cat_widget)

    def on_num_cat_widget_delete(self, widget: NumCatWidget):
        layout = self.numCatGUI.num_to_cat_layout
        ws = self.numCatGUI.num_to_cat_widgets

        if len(ws) > 1 and widget != ws[-1]:
            for widg in ws:
                layout.removeWidget(widg)
            ws.remove(widget)
            self.debug("Widget Removed!")
            for i, widg in enumerate(ws):
                layout.addWidget(widg, self._row(i), self._col(i))
            layout.update()
            self.debug("Layout Updated!")
        else:
            layout.removeWidget(widget)
            self.debug("Widget Removed!")
            _ = ws.pop()

        _ = self.numeric_to_categorical.pop(widget.column)
        self.replot_callbacks()

    def replot_callbacks(self):
        callbacks: dict[int, Callable[[], None]] = self.on_filter_change_callbacks
        for callback in callbacks.values():
            callback()
            self.debug("Replotting!")

    @cached_property
    def columns(self) -> list[str]:
        return [column for column in self.cleaned_data.columns]

    @cached_property
    def numeric_columns(self) -> list[str]:
        return [
            column
            for column in self.cleaned_data.columns
            if numeric_comparator(self.cleaned_data[column])
        ]

    @cached_property
    def categorical_columns(self) -> list[str]:
        return [
            column
            for column in self.cleaned_data.columns
            if categorical_comparator(self.cleaned_data[column])
        ]

    @cached_property
    def datetime_columns(self) -> list[str]:
        return [
            column
            for column in self.cleaned_data.columns
            if datetime_comparator(self.cleaned_data[column])
        ]

    @cache
    def get_column_dtype(self, column: str) -> Dtype:
        if numeric_comparator(self.cleaned_data[column]):
            return Dtype.NUMERIC
        elif categorical_comparator(self.cleaned_data[column]):
            return Dtype.CATEGORICAL
        else:
            return Dtype.DATETIME


@final
class DataModel:
    # Maps name of dataset to a DataStore
    datasets: dict[str, DataStore]
    active: str = ""

    def __init__(self, dataexplorer: "DataExplorer"):
        self.datasets = {}
        self.dataexplorer: "DataExplorer" = dataexplorer
        self.debug = dataexplorer.debug
        self.error = dataexplorer.error

    def add_dataset_to_model(
        self,
        dataset: pd.DataFrame,
        name: str,
        cleaning_operations: dict[str, dict[str, str]],
    ):
        self.debug(dataset.head())
        self.debug(dataset.shape)
        valid_name = name
        num = 1
        while valid_name in self.datasets:
            valid_name = f"{name} {num}"
            num += 1
        self.datasets[valid_name] = DataStore(
            cl_ops=cleaning_operations,
            data=dataset,
            name=valid_name,
            dataexplorer=self.dataexplorer,
        )
        self.dataexplorer.gui.add_filter_GUI_to_page(
            self.datasets[valid_name].filterGUI, valid_name
        )
        self.dataexplorer.gui.add_num_cat_GUI_to_page(
            self.datasets[valid_name].numCatGUI, valid_name
        )
        self.debug(f"Name used {valid_name}")
        self.dataexplorer.status_message(f"Data from {name} read!")
        self.dataexplorer.gui.update_data_list(valid_name)

    def set_active_dataset(self, name: str):
        self.active = name

    @property
    def active_dataset(self) -> DataStore | None:
        if self.active == "":
            self.error("No valid active dataset!")
            return None
        else:
            return self.datasets[self.active]

    @property
    def active_cleaned_data(self) -> pd.DataFrame | None:
        if self.active_dataset:
            return self.active_dataset.cleaned_data
        else:
            return None

    @property
    def active_filtered_data(self) -> pd.DataFrame | None:
        if self.active_dataset:
            return self.active_dataset.filtered_data
        else:
            return None


def handle_dtype_operation(
    series: pd.Series, operation: str, debug_callback: Callable[[str], None]
) -> pd.Series:
    debug_callback(operation)
    match operation:
        case DtypeOperation.CATEGORICAL.value:
            series = series.astype(str)
        case DtypeOperation.NUMERIC.value:
            series = pd.to_numeric(series, errors="coerce")  # pyright: ignore[reportUnknownMemberType, reportAssignmentType]
        case DtypeOperation.DATETIME.value:
            series = pd.to_datetime(series, errors="coerce")
        case _:
            raise RuntimeError("Invalid dtype cleaning operation.")

    assert isinstance(series, pd.Series)
    return series


def handle_nan_operation(
    df: pd.DataFrame,
    operation: str,
    column: str,
    debug_callback: Callable[[str], None],
) -> pd.DataFrame:
    debug_callback(operation)
    match operation:
        case NaNOperation.REPLACE_WITH_0.value:
            return df.fillna({column: 0})
        case NaNOperation.DROP_ROWS.value:
            return df.dropna(subset=column)
        case NaNOperation.KEEP_AS_NaN.value:
            return df
        case NaNOperation.REPLACE_WITH_NO_DATA.value:
            return df.fillna({column: "No Data"})
        case _:
            raise RuntimeError("Invalid nan cleaning operation.")


def numeric_comparator(df_column: pd.Series) -> bool:
    return is_numeric_dtype(df_column)


def categorical_comparator(df_column: pd.Series) -> bool:
    return is_object_dtype(df_column)


def datetime_comparator(df_column: pd.Series) -> bool:
    return is_datetime64_any_dtype(df_column)


def apply_filter(
    data: pd.DataFrame, column: str, filterstore: FilterStore
) -> pd.DataFrame:
    match filterstore.dtype:
        case Dtype.CATEGORICAL:
            if len(filterstore.filter_value) > 0:
                data = data[data[column].astype(str).isin(filterstore.filter_value)]
        case Dtype.NUMERIC:
            data = data[
                (data[column].values >= filterstore.filter_value[0])
                & (data[column].values <= filterstore.filter_value[1])
            ]
        case Dtype.DATETIME:
            data = data[
                (data[column].values >= filterstore.filter_value[0])
                & (data[column].values <= filterstore.filter_value[1])
            ]
    assert isinstance(data, pd.DataFrame)
    return data
